from __future__ import annotations
from typing import Literal
from .error import ParserException
from .path import Path, PathRange, ParsedPath, Redirect, Offset
from .token import Offset as TokenOffset
from .tokenizer import (
  EOF,
  Step,
  Symbol,
  Token,
  Tokenizer,
)


class _Parser:
  def __init__(self, content: str):
    self._cache_token: Token | None = None
    self._tokenizer: Tokenizer = Tokenizer(content)

  def parse(self) -> ParsedPath:
    paths = list(self._search_path())
    if len(paths) == 1:
      return paths[0]
    elif len(paths) == 3:
      # https://idpf.org/epub/linking/cfi/epub-cfi.html#sec-ranges
      parent, start, end = paths
      if parent.start_with_redirect():
        raise ParserException("Parent path cannot start with \"!\")")
      return PathRange(parent, start, end)
    else:
      raise ParserException(f"wrong path number: {len(paths)}")

  def _search_path(self):
    while True:
      path = self._parse_path()
      if path is None:
        raise ParserException(f"Unexpected token: {self._token}")

      yield path
      tail = self._token

      if isinstance(tail, Symbol) and tail.text == ",":
        self._forward()
      elif isinstance(tail, EOF):
        break
      else:
        raise ParserException(f"Unexpected token: {tail}")

  def _parse_path(self) -> Path | None:
    steps: list[Redirect | Step] = []
    offset: Offset | None = None

    while True:
      token = self._token
      if isinstance(token, Step):
        steps.append(token)
      elif isinstance(token, Symbol) and token.text == "!":
        if len(steps) > 0 and isinstance(steps[-1], Redirect):
          raise ParserException("Two consecutive redirects")
        steps.append(Redirect())
      else:
        break
      self._forward()

    if isinstance(self._token, TokenOffset):
      offset = self._token
      self._forward()

    if len(steps) == 0 and offset is None:
      return None

    if offset is None and isinstance(self._token, Symbol) and self._token.text == "!":
      # 参考 https://idpf.org/epub/linking/cfi/epub-cfi.html#sec-epubcfi-syntax
      # 从公式 redirected_path	=	"!" , ( offset | path ); 可知：
      # 不可以 ! 作为结尾，这样会选到某个文件的 root 节点（这是禁止的）
      raise ParserException("Cannot select root dom (must offset after redirect symbol)")

    return Path(steps=steps, offset=offset)

  def _is_symbol(self, symbol: Literal["!", ","]) -> bool:
    token = self._token
    if not isinstance(token, Symbol):
      return False
    return token.text == symbol

  @property
  def _token(self) -> Token:
    if self._cache_token is None:
      self._forward()
    return self._cache_token

  def _forward(self):
    if isinstance(self._cache_token, EOF):
      return
    while True:
      self._cache_token = self._tokenizer.read()
      if self._cache_token is not None:
        break

def parse(content: str) -> ParsedPath:
  return _Parser(content).parse()



import unittest

from epubcfi.tokenizer import Token, EOF, Tokenizer


class TestTokenizer(unittest.TestCase):

  def test_tokenizer(self):
    cfi_list = [
      "/6/4[chap01ref]!/4[body01]/10[para05]/3:10",
      "/6/4[chap^]01^^ref]!/4[body^[01]/10[para05]/3:10",
      "/6/4[chap01ref]@20:100",
      "/6/4[chap01ref]~2048",
      "/6/4[chap01ref]~2042@20:100",
    ]
    for cfi in cfi_list:
      token: Token
      tokens: list[Token] = []
      tokenizer = Tokenizer(cfi)

      while True:
        token = tokenizer.read()
        if isinstance(token, EOF):
          break
        tokens.append(token)

      self.assertEqual(cfi, "".join(map(str, tokens)))



import unittest

from epubcfi.cfi import parse, _capture_cfi

class TestHistoryBug(unittest.TestCase):

  def test_parse_and_compare(self):
    history_expressions = [
      "epubcfi(/6/24[id18]!/4/422/2/3,:227,:228)",
      "epubcfi(/6/22[id17]!/4/62/1,:0,:59)"
    ]
    for expression in history_expressions:
      _, cfi = _capture_cfi(expression)
      path = parse(expression)
      self.assertEqual(cfi, str(path))



from io import StringIO
from .error import TokenizerException


_escaped_chars = ("^", "[", "]", "(", ")", ",", ";", "=")

class _AssertionReader:
  def __init__(self, source: StringIO):
    self._source: StringIO = source
    self._buffer: StringIO = StringIO()
    self._escaped: bool = False

  def read(self) -> str:
    while True:
      char: str = self._source.read(1)
      assertion = self._read(char)
      if assertion is not None:
        return assertion

  def _read(self, char: str) -> str | None:
    if char == "":
      raise TokenizerException("Unexpected EOF")
    if self._escaped:
      if char not in _escaped_chars:
        raise TokenizerException(f"Unexpected character after escaped symbol: {char}")
      self._buffer.write(char)
      self._escaped = False
    elif char == "^":
      self._escaped = True
    elif char == "]":
      text = self._buffer.getvalue()
      if text == "":
        raise TokenizerException("Empty assertion is not allowed")
      return text
    else:
      self._buffer.write(char)
    return None

def read_assertion(source: StringIO) -> str:
  return _AssertionReader(source).read()

def str_assertion(assertion: str | None) -> str:
  if assertion is None:
    return ""
  buffer = StringIO()
  buffer.write("[")
  for char in assertion:
    if char in _escaped_chars:
      buffer.write("^")
    buffer.write(char)
  buffer.write("]")
  return buffer.getvalue()



from enum import Enum
from io import StringIO
from typing import Literal
from .assertion import read_assertion
from .token import Token, EOF, Symbol, Step, CharacterOffset, TemporalOffset, SpatialOffset, TemporalSpatialOffset
from .error import TokenizerException

class Phase(Enum):
  READY = 1
  STEP = 2
  OFFSET = 3

class Tokenizer:
  def __init__(self, content: str):
    self._phase: Phase = Phase.READY
    self._inject_char: str = ""
    self._buffer: StringIO = StringIO()
    self._source: StringIO = StringIO(content)
    self._offset_symbol: Literal[":", "@", "~"] = ":"
    self._offset_chain: list[tuple[Literal[":", "@", "~"], int]] = []

  def read(self) -> Token:
    char: str = ""
    while True:
      if self._inject_char == "":
        char = self._source.read(1)
      else:
        char = self._inject_char
        self._inject_char = ""

      token, forward = self._read(char)
      if not forward:
        self._inject_char = char
      if token is not None:
        return token

  def _read(self, char: str) -> tuple[Token | None, bool]:
    if self._phase == Phase.READY:
      return self._read_when_read(char)
    elif self._phase in (Phase.STEP, Phase.OFFSET):
      return self._read_integer(char)
    else:
      raise ValueError(f"Unexpected phase {self._phase}")

  def _read_when_read(self, char: str) -> tuple[Token | None, bool]:
    if char == "":
      return EOF(), True
    elif char in (",", "!"):
      return Symbol(text=char), True
    elif char == "/":
      self._phase = Phase.STEP
    elif char in (":", "~", "@"):
      self._offset_symbol = char
      self._phase = Phase.OFFSET
    else:
      raise TokenizerException(f"Unexpected character: {char}")
    return None, True

  def _read_integer(self, char: str) -> tuple[Token | None, bool]:
    if "0" <= char <= "9":
      self._buffer.write(char)
      return None, True
    else:
      text: str = self._buffer.getvalue()
      if len(text) > 1 and text.startswith("0"):
        raise TokenizerException(f"{text} leading zero is not allowed")
      integer = int(text)
      self._buffer = StringIO()

      if self._phase == Phase.STEP:
        assertion = self._read_assertion_if_need(char)
        step = Step(integer, assertion)
        self._phase = Phase.READY
        if assertion is None:
          return step, False
        else:
          return step, True

      elif self._phase == Phase.OFFSET:
        self._offset_chain.append((self._offset_symbol, integer))
        if char in (":", "~", "@"):
          self._offset_symbol = char
          return None, True
        assertion = self._read_assertion_if_need(char)
        offset = self._create_offset(assertion)
        self._phase = Phase.READY
        return offset, assertion is not None

      else:
        raise ValueError(f"Unexpected phase {self._phase}")

  def _read_assertion_if_need(self, char: str) -> str | None:
    assertion: str | None = None
    if char == "[":
      assertion = read_assertion(self._source)
    return assertion

  def _create_offset(self, assertion: str | None) -> Token:
    token: Token | None = None

    if len(self._offset_chain) == 1:
      symbol, value = self._offset_chain[0]
      if symbol == ":":
        token = CharacterOffset(
          value=value,
          assertion=assertion,
        )
      elif symbol == "~":
        token = TemporalOffset(
          seconds=value,
          assertion=assertion,
        )
    elif len(self._offset_chain) == 2:
      symbol1, value1 = self._offset_chain[0]
      symbol2, value2 = self._offset_chain[1]
      if symbol1 == "@" and symbol2 == ":":
        token = SpatialOffset(
          x=value1,
          y=value2,
          assertion=assertion,
        )
    elif len(self._offset_chain) == 3:
      symbol1, value1 = self._offset_chain[0]
      symbol2, value2 = self._offset_chain[1]
      symbol3, value3 = self._offset_chain[2]
      if symbol1 == "~" and symbol2 == "@" and symbol3 == ":":
        token = TemporalSpatialOffset(
          seconds=value1,
          x=value2,
          y=value3,
          assertion=assertion,
        )
    if token is None:
      raise TokenizerException(f"Unexpected offset: {self._str_offset_chain()}")
    self._offset_chain.clear()
    return token

  def _str_offset_chain(self) -> str:
    buffer = StringIO()
    for symbol, value in self._offset_chain:
      buffer.write(symbol)
      buffer.write(str(value))
    return buffer.getvalue()



import unittest

from epubcfi.path import PathRange
from epubcfi.parser import parse


class TestParser(unittest.TestCase):

  def test_parse_cfi(self):
    cfi_list = [
      "/6/4[chap01ref]!/4[body01]/10[para05]/3:10",
      "/6/4[chap^]01^^ref]!/4[body^[01]/10[para05]/3:10",
      "/6/4[chap01ref]@20:100",
      "/6/4[chap01ref]~2048",
      "/6/4[chap01ref]~2042@20:100",
    ]
    for cfi in cfi_list:
      path = parse(cfi)
      self.assertEqual(cfi, str(path))

  def test_simple_range(self):
    cfi_list = [
      "/6/4!/2[foobar],/10/4[foobar],/10",
      "/6/4!/2[foobar],!/10/4[foobar]:23,!/10",
      "/6/4!/2[foobar],!:12,!/10",
    ]
    for cfi in cfi_list:
      result = parse(cfi)
      self.assertTrue(isinstance(result, PathRange))
      self.assertEqual(cfi, str(result))



from dataclasses import dataclass
from functools import total_ordering
from typing import Any, Literal
from .assertion import str_assertion

@dataclass
class Token:
  def __str__(self) -> str:
    return ""

@dataclass
class EOF(Token):
  pass

@dataclass
class Symbol(Token):
  text: Literal[",", "!"]

  def __str__(self) -> str:
    return self.text

# https://idpf.org/epub/linking/cfi/epub-cfi.html#sec-sorting
@dataclass
@total_ordering
class Step(Token):
  index: int
  assertion: str | None

  def __str__(self) -> str:
    assertion = str_assertion(self.assertion)
    return f"/{self.index}{assertion}"

  def __lt__(self, obj: Any) -> bool:
    if not isinstance(obj, Step):
      return False
    return self.index < obj.index

  def __gt__(self, obj: Any) -> bool:
    if not isinstance(obj, Step):
      return True
    return self.index > obj.index

  def __le__(self, obj: Any) -> bool:
    if not isinstance(obj, Step):
      return False
    return self.index <= obj.index

  def __ge__(self, obj: Any) -> bool:
    if not isinstance(obj, Step):
      return True
    return self.index >= obj.index

  def __eq__(self, obj: Any) -> bool:
    if not isinstance(obj, Step):
      return False
    return self.index == obj.index

@dataclass
class Offset(Token):
  assertion: str | None

# https://idpf.org/epub/linking/cfi/epub-cfi.html#sec-path-terminating-char
@dataclass
@total_ordering
class CharacterOffset(Offset):
  value: int

  def __str__(self) -> str:
    assertion = str_assertion(self.assertion)
    return f":{self.value}{assertion}"

  def __lt__(self, obj: Any) -> bool:
    if not isinstance(obj, CharacterOffset):
      return False
    return self.value < obj.value

  def __gt__(self, obj: Any) -> bool:
    if not isinstance(obj, CharacterOffset):
      return True
    return self.value > obj.value

  def __le__(self, obj: Any) -> bool:
    if not isinstance(obj, CharacterOffset):
      return False
    return self.value <= obj.value

  def __ge__(self, obj: Any) -> bool:
    if not isinstance(obj, CharacterOffset):
      return True
    return self.value >= obj.value

  def __eq__(self, obj: Any) -> bool:
    if not isinstance(obj, CharacterOffset):
      return False
    return self.value == obj.value

# https://idpf.org/epub/linking/cfi/epub-cfi.html#sec-path-terminating-temporal
@dataclass
@total_ordering
class TemporalOffset(Offset):
  seconds: int

  def __str__(self) -> str:
    assertion = str_assertion(self.assertion)
    return f"~{self.seconds}{assertion}"

  def __lt__(self, obj: Any) -> bool:
    if not isinstance(obj, TemporalOffset):
      return False
    return self.seconds < obj.seconds

  def __gt__(self, obj: Any) -> bool:
    if not isinstance(obj, TemporalOffset):
      return True
    return self.seconds > obj.seconds

  def __le__(self, obj: Any) -> bool:
    if not isinstance(obj, TemporalOffset):
      return False
    return self.seconds <= obj.seconds

  def __ge__(self, obj: Any) -> bool:
    if not isinstance(obj, TemporalOffset):
      return True
    return self.seconds >= obj.seconds

  def __eq__(self, obj: Any) -> bool:
    if not isinstance(obj, TemporalOffset):
      return False
    return self.seconds == obj.seconds

# https://idpf.org/epub/linking/cfi/epub-cfi.html#sec-path-terminating-spatial
@dataclass
@total_ordering
class SpatialOffset(Offset):
  x: int
  y: int

  def __str__(self) -> str:
    assertion = str_assertion(self.assertion)
    return f"@{self.x}:{self.y}{assertion}"

  def __lt__(self, obj: Any) -> bool:
    if not isinstance(obj, SpatialOffset):
      return False
    return (self.y, self.x) < (obj.y, obj.x)

  def __gt__(self, obj: Any) -> bool:
    if not isinstance(obj, SpatialOffset):
      return True
    return (self.y, self.x) > (obj.y, obj.x)

  def __le__(self, obj: Any) -> bool:
    if not isinstance(obj, SpatialOffset):
      return False
    return (self.y, self.x) <= (obj.y, obj.x)

  def __ge__(self, obj: Any) -> bool:
    if not isinstance(obj, SpatialOffset):
      return True
    return (self.y, self.x) >= (obj.y, obj.x)

  def __eq__(self, obj: Any) -> bool:
    if not isinstance(obj, SpatialOffset):
      return False
    return (self.y, self.x) == (obj.y, obj.x)

# https://idpf.org/epub/linking/cfi/epub-cfi.html#sec-path-terminating-tempspatial
@dataclass
@total_ordering
class TemporalSpatialOffset(TemporalOffset):
  x: int
  y: int

  def __str__(self) -> str:
    assertion = str_assertion(self.assertion)
    return f"~{self.seconds}@{self.x}:{self.y}{assertion}"

  def __lt__(self, obj: Any) -> bool:
    if not isinstance(obj, TemporalSpatialOffset):
      return False
    return (self.seconds, self.y, self.x) < (obj.seconds, obj.y, obj.x)

  def __gt__(self, obj: Any) -> bool:
    if not isinstance(obj, TemporalSpatialOffset):
      return True
    return (self.seconds, self.y, self.x) > (obj.seconds, obj.y, obj.x)

  def __le__(self, obj: Any) -> bool:
    if not isinstance(obj, TemporalSpatialOffset):
      return False
    return (self.seconds, self.y, self.x) <= (obj.seconds, obj.y, obj.x)

  def __ge__(self, obj: Any) -> bool:
    if not isinstance(obj, TemporalSpatialOffset):
      return True
    return (self.seconds, self.y, self.x) >= (obj.seconds, obj.y, obj.x)

  def __eq__(self, obj: Any) -> bool:
    if not isinstance(obj, TemporalSpatialOffset):
      return False
    return (self.seconds, self.y, self.x) == (obj.seconds, obj.y, obj.x)



import re

from .parser import parse as parse_cfi
from .path import Path, PathRange, ParsedPath


def parse(path: str) -> ParsedPath | None:
  _, cfi = _capture_cfi(path)
  if cfi is None:
    return path, None
  return parse_cfi(cfi)

def split(path: str) -> tuple[str, ParsedPath | None]:
  tail, cfi = _capture_cfi(path)
  if cfi is None:
    return path, None
  result = parse_cfi(cfi)
  prefix = path[:len(path) - len(tail)]
  return prefix, result

def to_absolute(r: PathRange) -> tuple[Path, Path]:
  start = Path(
    steps=r.parent.steps + r.start.steps,
    offset=r.start.offset,
  )
  end = Path(
    steps=r.parent.steps + r.end.steps,
    offset=r.end.offset,
  )
  return start, end

def _capture_cfi(path: str):
  matched = re.search(r"(#|^)epubcfi\((.*)\)$", path)
  if matched:
    return matched.group(), matched.group(2)
  else:
    return None, None



# https://idpf.org/epub/linking/cfi/epub-cfi.html#sec-sorting

from __future__ import annotations
from io import StringIO
from dataclasses import dataclass
from functools import total_ordering
from typing import Any
from .tokenizer import (
  Step,
  CharacterOffset,
  TemporalOffset,
  SpatialOffset,
  TemporalSpatialOffset,
)

Offset = CharacterOffset | TemporalOffset | SpatialOffset | TemporalSpatialOffset

@dataclass
@total_ordering
class Redirect:
  def __str__(self) -> str:
    return "!"

  def __lt__(self, _: Any) -> bool:
    return False

  def __gt__(self, _: Any) -> bool:
    return False

  def __le__(self, obj: Any) -> bool:
    return isinstance(obj, Redirect)

  def __ge__(self, obj: Any) -> bool:
    return isinstance(obj, Redirect)

  def __eq__(self, obj: Any) -> bool:
    return isinstance(obj, Redirect)

@dataclass
@total_ordering
class Path:
  steps: list[Redirect | Step]
  offset: Offset | None

  def start_with_redirect(self) -> bool:
    return isinstance(self.steps[0], Redirect)

  def __str__(self):
    buffer = StringIO()
    for step in self.steps:
      buffer.write(str(step))
    if self.offset is not None:
      buffer.write(str(self.offset))
    return buffer.getvalue()

  def __lt__(self, obj: Any) -> bool:
    if not isinstance(obj, ParsedPath):
      return True
    tail1, tail2 = self._skip_common_steps_head(obj)
    return tail1 < tail2

  def __gt__(self, obj: Any) -> bool:
    if not isinstance(obj, ParsedPath):
      return False
    tail1, tail2 = self._skip_common_steps_head(obj)
    return tail1 > tail2

  def __le__(self, obj: Any) -> bool:
    if not isinstance(obj, ParsedPath):
      return True
    tail1, tail2 = self._skip_common_steps_head(obj)
    return tail1 <= tail2

  def __ge__(self, obj: Any) -> bool:
    if not isinstance(obj, ParsedPath):
      return False
    tail1, tail2 = self._skip_common_steps_head(obj)
    return tail1 >= tail2

  def __eq__(self, obj: Any) -> bool:
    if not isinstance(obj, ParsedPath):
      return False
    tail1, tail2 = self._skip_common_steps_head(obj)
    return tail1 == tail2

  def _skip_common_steps_head(self, obj: ParsedPath):
    obj_steps: list[Redirect | Step]
    obj_offset: Offset | None
    index: int = 0

    if isinstance(obj, PathRange):
      obj_steps = obj.parent.steps + obj.start.steps
      obj_offset = obj.start.offset
    else:
      obj_steps = obj.steps
      obj_offset = obj.offset

    for s1, s2 in zip(self.steps, obj_steps):
      if s1 != s2:
        break
      index += 1

    tail1: Redirect | Step | Offset | None = None
    tail2: Redirect | Step | Offset | None = None

    if index < len(self.steps):
      tail1 = self.steps[index]
    else:
      tail1 = self.offset
    if index < len(obj_steps):
      tail2 = obj_steps[index]
    else:
      tail2 = obj_offset

    type1 = self._offset_type_id(tail1)
    type2 = self._offset_type_id(tail2)

    if type1 < type2:
      return (0, 1)
    elif type1 > type2:
      return (1, 0)
    else:
      return (tail1, tail2)

  def _offset_type_id(self, tail: Redirect | Step | Offset | None):
    # https://idpf.org/epub/linking/cfi/epub-cfi.html#sec-sorting
    # different step types come in the following order from least important to most important:
    # character offset (:), child (/), temporal-spatial (~ or @), reference/indirect (!).
    if tail is None:
      return 0
    elif isinstance(tail, Redirect):
      return 1
    elif isinstance(tail, SpatialOffset):
      return 2
    elif isinstance(tail, TemporalOffset):
      return 3
    elif isinstance(tail, TemporalSpatialOffset):
      return 4
    elif isinstance(tail, Step):
      return 5
    elif isinstance(tail, CharacterOffset):
      return 6
    else:
      raise ValueError(f"Unknown offset type: {tail}")

@dataclass
@total_ordering
class PathRange:
  parent: Path
  start: Path
  end: Path

  def __str__(self):
    return f"{self.parent},{self.start},{self.end}"

  def __lt__(self, obj: Any) -> bool:
    if not isinstance(obj, ParsedPath):
      return True
    if isinstance(obj, Path):
      return obj > self
    return self._to_tuple() < obj._to_tuple()

  def __gt__(self, obj: Any) -> bool:
    if not isinstance(obj, ParsedPath):
      return False
    if isinstance(obj, Path):
      return obj < self
    return self._to_tuple() > obj._to_tuple()

  def __le__(self, obj: Any) -> bool:
    if not isinstance(obj, ParsedPath):
      return True
    if isinstance(obj, Path):
      return obj >= self
    return self._to_tuple() <= obj._to_tuple()

  def __ge__(self, obj: Any) -> bool:
    if not isinstance(obj, ParsedPath):
      return False
    if isinstance(obj, Path):
      return obj <= self
    return self._to_tuple() >= obj._to_tuple()

  def __eq__(self, obj: Any) -> bool:
    if not isinstance(obj, ParsedPath):
      return False
    if isinstance(obj, Path):
      return obj == self
    return self._to_tuple() == obj._to_tuple()

  def _to_tuple(self):
    return (self.parent, self.start, self.end)

  def _obj_to_tuple(self, obj: ParsedPath):
    if isinstance(obj, PathRange):
      return (obj.parent, obj.start, obj.end)
    else:
      return obj, obj, obj

ParsedPath = Path | PathRange



from setuptools import setup

setup(
  name="epubcfi",
  version="0.0.1",
  author="Tao Zeyu",
  author_email="i@taozeyu.com",
  description="handle Epub CFI",
  packages=["epubcfi"],
  long_description=open("./README.md", encoding="utf8").read(),
  long_description_content_type="text/markdown",
  install_requires=[],
)


import unittest

from epubcfi.path import PathRange
from epubcfi.cfi import parse, split, to_absolute, _capture_cfi


class TestCFI(unittest.TestCase):

  def test_to_absolute(self):
    prefix, results = split("book.epub#epubcfi(/6/4,!/2[foobar],/10/4[foz])")
    self.assertTrue(isinstance(results, PathRange))
    start, end = to_absolute(results)
    self.assertEqual(prefix, "book.epub")
    self.assertEqual(str(start), "/6/4!/2[foobar]")
    self.assertEqual(str(end), "/6/4/10/4[foz]")

  def test_format(self):
    prefix, results = split("book.epub#epubcfi(/6/4,!/2[foobar],/10/4[foz])")
    self.assertTrue(isinstance(results, PathRange))
    self.assertEqual(prefix, "book.epub")
    self.assertEqual(str(results), "/6/4,!/2[foobar],/10/4[foz]")

  def test_capture_cfi(self):
    pairs = [(
      "book.epub#epubcfi(/6/4[chap01ref]!/4[body01]/10[para05]/3:10)",
      "/6/4[chap01ref]!/4[body01]/10[para05]/3:10",
    ), (
      "epubcfi(/6/4[chap01ref]!/4[body01]/10[para05]/3:10)",
      "/6/4[chap01ref]!/4[body01]/10[para05]/3:10",
    ), (
      "epubcfi(/6/4[chap01ref]!/4[body01]/10[para05]/3:10)foobar",
      None,
    )]
    for source, target in pairs:
      _, cfi = _capture_cfi(source)
      self.assertEqual(cfi, target)

  def test_sorted_cfi(self):
    expressions = [
      "book.epub#epubcfi(/6/4/321)",
      "book.epub#epubcfi(/6/4)",
      "book.epub#epubcfi(/6/4/123)",
      "book.epub#epubcfi(/6/4@0:99)",
      "book.epub#epubcfi(/6/4@100:33)",
      "book.epub#epubcfi(/6/4:23)",
    ]
    cfi_list = [parse(e) for e in expressions]
    cfi_list = [cfi for cfi in cfi_list if cfi is not None]
    cfi_list.sort()
    self.assertEqual([str(cfi) for cfi in cfi_list], [
      "/6/4",
      "/6/4@100:33",
      "/6/4@0:99",
      "/6/4/123",
      "/6/4/321",
      "/6/4:23",
    ])



from .cfi import parse, split, to_absolute
from .path import Path, PathRange, ParsedPath, Offset, Redirect
from .token import Offset as BaseOffset
from .tokenizer import Step, CharacterOffset, TemporalOffset, SpatialOffset, TemporalSpatialOffset
from .error import ParserException, TokenizerException, EpubCFIException
